/*

Arquivo:

   Item.cpp

Descrio:

   Implementao das classes em Item.hpp.

*/

/* Especfico do gcc */
#pragma implementation
/* Fim do especfico */

//
// Cabealhos
//
#include <stdio.h>
#include <string.h>
#include "Item.hpp"

//
// Implementao - Item
//

/*
 * Item(char*, mode = copy_pointer, unsigned long = 1)
 *   Inicializa a classe com uma palavra e uma determinada frequncia.
 *
 * Argumentos:
 * [in]  src_word             Ponteiro para uma sequncia de caracteres que re-
 *                            presenta a palavra a ser includa.
 * [opc] cpy_mode             Se igual a copy_pointer, apenas o ponteiro 
 *                            src_word  copiado. Se igual a copy_string, 
 *                            novo espao  alocado e a string copiada.
 * [opc] start_freq           Frequncia inicial da palavra
 *
 * Excees:
 *   Se falhar em alocar memria, causa AllocationError. Se a chave especifica-
 * da for vazia, causa EmptyKey
 *
 */
Item::Item(char* src_word, mode cpy_mode = copy_pointer,
	   unsigned long start_freq = 1)
{
  if(src_word == NULL)
    throw EmptyKey();

  /* Ajuste da frequncia */
  freq = start_freq;

  /* Ajuste da palavra */
  if(cpy_mode == copy_pointer) {
    word = src_word;
    return;
  }

  word = new char[strlen(src_word) + 1];

  if(word == NULL)
    throw AllocationError();

  strcpy(word, src_word);
}

//
// Implementao - STItem
//

/*
 * Este ponteiro contm o endereo da funo de comparao que  normalmente
 * executada. Pode ser modificado atravs de SETcompare_mode
 *
 */
int (*STItem::cmp_func)(const Item&, const Item&) = &STItem::ITEMcompare_word;

/*
 * KEYCompare
 *   Verifica se um elemento possui determinada chave.
 *
 * Argumentos:
 * [in]  obj                  Elemento a ser verificado
 *       key                  Chave a ser comparada
 *
 * Excees:
 *   Causa a exceo EmptyKey se key == NULL
 *
 * Retornos:
 *   Retorna true se a chave do elemento e key forem iguais, false caso contr-
 * rio.
 *
 */
bool STItem::KEYcompare(const Item& obj, const char* key)
{
  if(key == NULL)
    throw Item::EmptyKey();

  return (strcmp(obj.word, key) == 0);
}

/*
 * KEYhash
 *   Calcula uma funo de hash para uma string. Utiliza universal hash.
 *
 * Argumentos:
 * [in]  key                  Chave para a qual se vai calcular a funo
 *       m                    Tamanho da tabela. O retorno ser um inteiro
 *                            entre 0 e m-1, inclusive
 * 
 * Retornos:
 *   Um inteiro entre 0 e m-1, inclusive
 *
 * Referncias:
 *   Algorithms in C, parts 1-4, Robert Sedgewick, Captulo 14 - pg. 579
 *
 */
unsigned long STItem::KEYhash(const char* key, unsigned long m)
{
  unsigned long h;
  int a = 31415, b = 27183;

  for(h = 0; *key != '\0'; key++, a = a*b % (m-1))
    h = (a*h + (unsigned char) *key) % m;

  return h;
}

/*
 * SETcompare_mode
 *   Especifica mtodo de comparao entre objetos, alterando a varivel
 * esttica cmp_func.
 *
 * Argumentos:
 * [in]  m                    Modo de comparao: compare_word ou compare_freq
 *
 */
void STItem::SETcompare_mode(cmp_mode m)
{
  if(m == compare_word) 
    cmp_func = &ITEMcompare_word;
  else
    cmp_func = &ITEMcompare_freq;
}

/*
 * ITEMcompare_freq
 *   Compara dois itens diferentes segundo a frequncia.
 *
 * Argumentos:
 * [in]  obj1                 Primeiro elemento
 *       obj2                 Segundo elemento
 *
 * Retornos:
 *   Se obj1 < obj2 retorna greater, se obj1 = obj2 retorna equal e, se 
 * obj1 > obj2 retorna lesser. Os retornos so assim para que as palavras mais
 * frequentes sejam colocadas a frente das menos frequentes.
 *
 */
int STItem::ITEMcompare_freq(const Item& obj1, const Item& obj2)
{
  if(obj1.freq > obj2.freq)
    return lesser;
  else if(obj1.freq == obj2.freq)
    return equal;
  else
    return greater;
}

/*
 * ITEMcompare_word
 *   Compara dois itens diferentes segundo a palavra.
 *
 * Argumentos:
 * [in]  obj1                 Primeiro elemento
 *       obj2                 Segundo elemento
 *
 * Retornos:
 *   Se obj1 > obj2 retorna greater, se obj1 = obj2 retorna equal e, se 
 * obj1 < obj2 retorna lesser.
 *
 */
int STItem::ITEMcompare_word(const Item& obj1, const Item& obj2)
{
  int ret = strcmp(obj1.word, obj2.word);

  if(ret > 0)
    return greater;
  else if(ret == 0)
    return equal;
  else
  return lesser;
}
