/*

Arquivo:

   ST.cpp

Descrio:

   Implementao da tabela de smbolos, cuja interface se encontra em ST.hpp.
 Usa uma tabela de hash, de tamanho primo, com separate chaining, que aumenta
 de tamanho no decorrer do programa.

Referncias:

   Ver Algorithms in C, Parts 1-4, Robert Sedgewick, Captulo 14
*/

/* Especfico do gcc */
#pragma implementation
/* Fim do especfico */

//
// Cabealhos
//
#include <stdio.h>
#include <stdlib.h>
#include "ST.hpp"
#include "Item.hpp"

/* Especfico do gcc */
typedef STable<Item, char*, STItem> __STABLE_ITEM_CHAR_STITEM__;
/* Fim do especfico */

//
// Constantes
//
static const unsigned long primes[] = { 1021, 2039, 4093, 8191, 16381,
					32749, 65521,131071, 262139, 524287, 
					1048573, 2097143, 4194301, 9388593,
					16777213, 3554393, 67108859, 134217689, 
					268435399, 536870909, 1073741789,
					2147483647, 0 };

//
// Implementao
//

/* Funes de linkagem interna */
template<class IT, class KEY, class CMP>
static int compare(const void*, const void*);

/* Esta funo  utilizada apenas internamente, numa chamada a funo qsort */
template<class IT, class KEY, class CMP>
static int compare(const void* obj1, const void* obj2)
{
  int ret = CMP::ITEMcompare(**((Item**) obj1), **((Item**) obj2));

  if(ret == CMP::lesser)
    return -1;
  else if(ret == CMP::equal)
    return 0;
  else
    return 1;
}

/* Funes Privadas */

/*
 * DoubleTableSize
 *   Dobra o tamanho da tabela de hashing, re-inserindo todos os elementos.
 *
 * Excees:
 *   Se, por algum motivo, falha em criar uma tabela com o dobro do tamanho,
 * mantm a atual e causa a exceo DoubleTableSizeError.
 *
 */
template<class IT, class KEY, class CMP>
void STable<IT, KEY, CMP>::DoubleTableSize()
{
  int new_size = table_size + 1;

  if(primes[new_size] == 0)
    return;

  /* Cria uma nova tabela de hashing */
  Node** temp_table = new (Node*)[primes[new_size]];

  if(temp_table == NULL)
    throw DoubleTableSizeError();

  for(unsigned long i = 0; i < primes[new_size]; temp_table[i++] = NULL);

  /* Re-insere todos os elementos */
  for(unsigned long i = 0; i < primes[table_size]; i++) {
    if(hash_table[i] == NULL)
      continue;

    for(Node* curr = hash_table[i], *tmp; curr != NULL; curr = tmp) {
      unsigned long dest = CMP::KEYhash(*(curr->data), primes[new_size]);

      tmp = curr->next;
      curr->next = temp_table[dest];
      temp_table[dest] = curr;
    }
  }

  /* Elimina tabela antiga */
  delete[] hash_table;

  table_size++;
  hash_table = temp_table;
}

/* Construtores e Destrutores */

/*
 * STable
 *   Constri uma tabela de hashing vazia com primes[0] posies
 *
 * Excees:
 *   Causa AllocationError se falhar em alocar a tabela.
 *
 */
template<class IT, class KEY, class CMP>
STable<IT, KEY, CMP>::STable()
{
  item_count = 0;
  sorted_table = NULL;

  /* Tamanho inicial da tabela, como ndice do array primes */
  table_size = 0;

  /* Alocao da tabela de hashing */
  hash_table = new (Node*)[primes[table_size]];

  if(hash_table == NULL)
    throw AllocationError();

  /* Inicializao da tabela */
  for(unsigned long i = 0; i < primes[table_size]; hash_table[i++] = NULL);
}

/*
 * ~STable
 *   Destri todos os elementos inseridos na tabela e a tabela de hashing.
 *
 */
template<class IT, class KEY, class CMP>
STable<IT, KEY, CMP>::~STable()
{
  for(unsigned long i = 0; i < primes[table_size]; i++) {
    if(hash_table[i] == NULL)
      continue;

    for(Node* curr = hash_table[i], *tmp; curr != NULL; curr = tmp) {
      tmp = curr->next;
      delete curr->data;
      delete curr;
    }
  }

  delete[] hash_table;
  delete[] sorted_table;
}

/* Mtodos */

/*
 * insert
 *   Insere um elemento na tabela de smbolos, dado seu ponteiro
 *
 * Argumentos:
 * [in]  obj                  Ponteiro para o elemento a ser inserido
 *
 * Excees:
 *   Causa AllocationError se falhar em alocar memria para um novo n.
 *
 */
template<class IT, class KEY, class CMP>
void STable<IT, KEY, CMP>::insert(IT* obj)
{
  /* Se um array foi construdo, elimin-lo */
  if(sorted_table != NULL) {
    delete[] sorted_table;
    sorted_table = NULL;
  }

  /* Verifica se h necessidade de dobrar o tamanho da tabela
   * O critrio utilizado  o nmero de elementos ser mais do que 10 vezes maior
   * que o tamanho da tabela e hashing, ou seja, h em mdia 10 elementos por
   * categoria da tabela
   */
  if(item_count > 10 * primes[table_size])
    DoubleTableSize();

  /* Computa funo de hash */
  unsigned long dest = CMP::KEYhash(*obj, primes[table_size]);

  /* Cria o novo n */
  Node* new_node = new Node;

  if(new_node == NULL)
    throw AllocationError();

  /* Insere o n */
  new_node->data = obj;
  new_node->next = hash_table[dest];
  hash_table[dest] = new_node;

  item_count++;
}

/*
 * search
 *   Procura por um elemento que possua determinada chave.
 *
 * Argumentos:
 * [in]  key                  Chave
 *
 * Retornos:
 *   Retorna um ponteiro para o elemento se este for encontrado ou NULL caso
 * contrrio.
 *
 */
template<class IT, class KEY, class CMP>
IT* STable<IT, KEY, CMP>::search(const KEY& key) const
{
  /* Computa a funo de hash */
  unsigned long dest = CMP::KEYhash(key, primes[table_size]);

  /* Procura o elemento na tabela */
  Node* curr = hash_table[dest];
  for( ; curr != NULL; curr = curr->next)
    if(CMP::KEYcompare(*(curr->data), key))
      break;

  if(curr == NULL)
    return NULL;
  else
    return curr->data;
}

/*
 * build_array
 *   Constri um array com um elemento para cada item inserido na tabela.
 * Se um array j houver sido criado, no executa nenhuma operao.
 *
 * Excees:
 *   Causa a exceo AllocationError se falhar em alocar memria para o
 * array.
 *
 */
template<class IT, class KEY, class CMP>
void STable<IT, KEY, CMP>::build_array()
{
  if(sorted_table != NULL)
    return;

  /* Cria vetor para guardar os elementos */
  sorted_table = new (IT*)[item_count];

  if(sorted_table == NULL)
    throw AllocationError();

  /* Atribui a cada elemento do vetor um ponteiro para um dos itens */
  for(unsigned long i = 0, p = 0; i < primes[table_size]; i++) {
    if(hash_table[i] == NULL)
      continue;

    for(Node* curr = hash_table[i]; curr != NULL; curr = curr->next, p++)
      sorted_table[p] = curr->data;
  }
}

/*
 * sort
 *   Ordena a tabela de smbolos. Usa a funo qsort. Um array para guardar
 * os elementos da tabela deve ter sido construdo antes desta funo ser  
 * utilizada.
 *
 * Argumentos:
 * [opc]  start               ndice a partir do qual se deve ordenar.
 *                            Default  0.
 * [opc]  num_items           Nmero de itens a serem ordenados, a partir
 *                            de start. Se igual a 0, item_count - start
 *                            itens sero ordenados. Default  0.
 *
 * Excees:
 *   Se um array no tiver sido construdo, causa ArrayNotBuilt.
 *
 */
template<class IT, class KEY, class CMP>
void STable<IT, KEY, CMP>::sort(unsigned long start = 0, 
				unsigned long num_items = 0)
{
  if(item_count == 0)
    return;

  if(sorted_table == NULL)
    throw ArrayNotBuilt();

  if(num_items == 0)
    num_items = item_count - start;

  /* Chama a funo qsort */
  qsort(sorted_table + start, num_items, sizeof(IT*), &compare<IT, KEY, CMP>);
}

//
// Operadores
//

/*
 * Operador []
 *   Utilizado para iterar pela tabela quando foi construdo um array.
 *
 * Argumentos:
 * [in]  i                    Posio a ser retornada
 *
 * Retornos:
 *   Retorna uma referncia para o item da tabela.
 *
 * Excees:
 *   Causa ArrayNotBuilt se um array no tiver sido construdo.
 *
 */
template<class IT, class KEY, class CMP>
IT& STable<IT, KEY, CMP>::operator[](unsigned long i) const
{
  if(sorted_table == NULL)
    throw ArrayNotBuilt();

  return *sorted_table[i];
}
